"""Tests for canalweb.vod."""

# pylint: disable-msg=C0103,W0704,W0212,W0201

import os
import shutil

import canalweb
from canalweb import vod

ROOT = os.path.dirname(__file__)


class TestDownloader(object):
    """Test the Downloader class."""

    def setup_method(self, _):
        """Setup a downloader."""
        self.cache = os.path.join(ROOT, 'test-cache')
        try:
            shutil.rmtree(self.cache)
        except OSError:
            pass
        os.mkdir(self.cache)

        self.dl = vod.Downloader(self.cache)

    def test_content_ids(self):
        """Check exraction of video IDs."""
        fd = open(os.path.join(ROOT, 'testdata', 'pid1830-c-zapping.html'))

        assert [url for url in self.dl._extract_content_ids(fd)] == [
            195190, 195164, 194541, 194196, 193817, 193423, 193128, 192859]

    def test_config(self):
        """Check parsing of the XML file."""
        data = open(os.path.join(ROOT, 'testdata', 'video-config.xml')).read()
        show = canalweb.SHOW_BY_NICK['guignoles']
        video = self.dl._video_url(data, show, 1234)

        assert video.url == (
            'http://vod-flash.canalplus.fr/WWWPLUS/PROGRESSIF'
            '/0812/LES_GUIGNOLS_EMISSION_081214_AUTO_1138_169_video_H.flv')



from struct import unpack
from datetime import datetime

class FLVReader(dict):
    """
    Reads metadata from FLV files
    """

    # Tag types
    AUDIO = 8
    VIDEO = 9
    META = 18
    UNDEFINED = 0

    def __init__(self, filename):
        """return a dictionary of meta data."""
        # Lock on to the file
        self.file = open(filename, 'rb')
        self.signature = self.file.read(3)
        assert self.signature == 'FLV', 'Not an flv file'
        self.version = self.readbyte()
        self.typeFlags = self.readbyte()
        self.dataOffset = self.readint()
        extraDataLen = self.dataOffset - self.file.tell()
        self.extraData = self.file.read(extraDataLen)
        self.readtag()

    def readtag(self):
        unknown = self.readint()
        tagType = self.readbyte()
        dataSize = self.read24bit()
        timeStamp = self.read24bit()
        unknown = self.readint()
        if tagType == self.AUDIO:
            print "Can't handle audio tags yet"
        elif tagType == self.VIDEO:
            print "Can't handle video tags yet"
        elif tagType == self.META:
            endpos = self.file.tell() + dataSize
            event = self.readAMFData()
            metaData = self.readAMFData()
            # We got the meta data.
            # Our job is done.
            # We are complete
            self.update(metaData)
        elif tagType == self.UNDEFINED:
            print "Can't handle undefined tags yet"

    def readint(self):
      data = self.file.read(4)
      return unpack('>I', data)[0]

    def readshort(self):
      data = self.file.read(2)
      return unpack('>H', data)[0]

    def readbyte(self):
      data = self.file.read(1)
      return unpack('B', data)[0]

    def read24bit(self):
      b1, b2, b3 = unpack('3B', self.file.read(3))
      return (b1 << 16) + (b2 << 8) + b3

    def readAMFData(self, dataType=None):
        if dataType is None:
            dataType = self.readbyte()
        funcs = {
            0: self.readAMFDouble,
            1: self.readAMFBoolean,
            2: self.readAMFString,
            3: self.readAMFObject,
            8: self.readAMFMixedArray,
           10: self.readAMFArray,
           11: self.readAMFDate
        }
        func = funcs[dataType]
        if callable(func):
            return func()

    def readAMFDouble(self):
        return unpack('>d', self.file.read(8))[0]

    def readAMFBoolean(self):
        return self.readbyte() == 1

    def readAMFString(self):
        size = self.readshort()
        return self.file.read(size)

    def readAMFObject(self):
        data = self.readAMFMixedArray()
        result = object()
        result.__dict__.update(data)
        return result

    def readAMFMixedArray(self):
        size = self.readint()
        result = {}
        for i in range(size):
            key = self.readAMFString()
            dataType = self.readbyte()
            if not key and dataType == 9:
                break
            result[key] = self.readAMFData(dataType)
        return result

    def readAMFArray(self):
        size = self.readint()
        result = []
        for i in range(size):
            result.append(self.readAMFData)
        return result

    def readAMFDate(self):
        return datetime.fromtimestamp(self.readAMFDouble())
        

if __name__ == '__main__':
    import sys
    from pprint import pprint
    if len(sys.argv) == 1:
        print 'Usage: %s filename [filename]...' % sys.argv[0]
        print 'Where filename is a .flv file'
        print 'eg. %s myfile.flv' % sys.argv[0]
    for fn in sys.argv[1:]:
        x = FLVReader(fn)
        pprint(x)



import datetime

import canalweb
from canalweb import vod
from canalweb import xspf


def test_format_xspf():
    now = datetime.date(2008, 12, 12)

    videos = [
        vod.VOD('http://foo/bar', canalweb.SHOW_BY_NICK['zapping'], 456, now),
        vod.VOD('http://foo/baz', canalweb.SHOW_BY_NICK['zapping'], 457, now)
    ]
    videos[1].local_path = '/to/the/file'

    data = xspf.videos_to_xspf(videos)
    print data
    assert data == '''\
<?xml version="1.0" encoding="UTF-8"?>
<playlist version="1" xmlns="http://xspf.org/ns/0/">
  <title>Canal Web</title>
  <trackList>
    <track>
      <location>http://foo/bar</location>
      <title>Le Zapping 2008-12-12 (456)</title>
    </track>
    <track>
      <location>file:///to/the/file</location>
      <title>Le Zapping 2008-12-12 (457)</title>
    </track>
  </trackList>
</playlist>
'''
    



"""Commandline interface for managing the videos."""

import logging
import optparse
import os
import sys
import ConfigParser

import canalweb

from canalweb import vod
from canalweb import xspf


def get_config(marker):
    """Parse the config from the specified file."""
    parser = ConfigParser.SafeConfigParser()
    parser.readfp(open(marker))

    if not parser.has_section('seen'):
        parser.add_section('seen')
    return parser


def set_config(marker, parser):
    """Store the updated config."""
    parser.write(open(marker, 'w'))


def main(args=None):
    """Main entry point for the script."""
    if args is None:
        args = sys.argv[1:]

    parser = optparse.OptionParser(
        usage=('%prog [options] init | sync <show...> | seen <show...>\n\n' +
               'Available shows: %s''' % (
        ' '.join(sorted(canalweb.SHOW_BY_NICK)),)))
    parser.add_option('-c', '--cache',
                      action='store', dest='cache', default='.',
                      help='Directory in which to download files')

    (options, args) = parser.parse_args(args)

    # 'init': create the CANALWEB file
    # 'sync': download all files
    # 'list': list the available files
    if not args:
        parser.error('missing command')
    command = args.pop(0)

    logging.basicConfig(level=logging.DEBUG)

    cache = os.path.abspath(options.cache)
    marker = os.path.join(cache, 'CANALWEBCACHE')

    if command == 'init':
        logging.info('initializing new cache')
        if os.path.exists(marker):
            logging.error('directory already initialized')
            return 1
        open(marker, 'w').close()
        return 0

    # sanity check: have a predefined file in the cache directory
    if not os.path.exists(marker):
        logging.error('invalid cache location %r: initialize it first' % (
            cache,))
        return 1

    if command not in ('sync', 'seen'):
        parser.error('unknown command %r' % (command,))

    if not args:
        parser.error('missing show names')

    try:
        shows = [canalweb.SHOW_BY_NICK[nick] for nick in args]
    except KeyError, error:
        parser.error('unknown show %r' % (error.args,))
        
    downloader = vod.Downloader(cache)
    all_videos = downloader.get_cached_configs()
    parser = get_config(marker)

    if command == 'sync':
        for show in shows:
            # these are all the videos available right now
            videos = all_videos.get(show, set())
            videos.update(downloader.get_video_info(show))

            if parser.has_option('seen', show.nickname):
                last_seen = parser.getint('seen', show.nickname)
            else:
                last_seen = 0

            current_videos = sorted((video for video in videos
                                     if video.vid > last_seen),
                                    key=lambda x: x.vid)
            to_delete = [video for video in videos if video.vid <= last_seen]
            for video in to_delete:
                logging.info('%r can be deleted' % (video,))
                downloader.delete_video(video)
            # all the others can be deleted
            for video in current_videos:
                downloader.fetch_video(video)
            # remove old videos that are not in the list
            playlist = '%s.xspf' % (show.nickname,)
            open(playlist, 'w').write(xspf.videos_to_xspf(current_videos))
            logging.info('generated %s' % (playlist,))

    elif command == 'seen':

        for show in shows:
            latest = max(v.vid for v in all_videos[show])
            parser.set('seen', show.nickname, str(latest))

        set_config(marker, parser)
        



"""Manage videos from the Canal+ website."""

class Show(object):
    """Describe a single Canal+ show."""
    # pylint: disable-msg=R0903
    def __init__(self, pid, nickname, fullname):
        self.pid = pid
        self.nickname = nickname
        self.fullname = fullname


SHOW_BY_NICK = {}
SHOW_BY_PID = {}

for show in [Show(1830, 'zapping', u'Le Zapping'),
             Show(1784, 'guignoles', u'Les Gignoles'),
             Show(1787, 'magzin', u'Groland Magzin')]:
    SHOW_BY_NICK[show.nickname] = show
    SHOW_BY_PID[show.pid] = show



"""Generate a playlist in XSPF format."""

from xml.sax import saxutils

def videos_to_xspf(videos):
    tracks = []
    for video in videos:
        if video.local_path:
            url = 'file://' + video.local_path
        else:
            url = video.url
        url = saxutils.escape(url)

        title = u'%s %s (%d)' % (
            video.show.fullname, video.nice_date(), video.vid)
        title = saxutils.escape(title.encode('utf-8'))

        tracks.append('''\
    <track>
      <location>%(location)s</location>
      <title>%(title)s</title>
    </track>''' % {'title': title, 'location': url})
        
    return '''\
<?xml version="1.0" encoding="UTF-8"?>
<playlist version="1" xmlns="http://xspf.org/ns/0/">
  <title>Canal Web</title>
  <trackList>
%s
  </trackList>
</playlist>
''' % ('\n'.join(tracks),)



import ez_setup
ez_setup.use_setuptools()

from setuptools import setup, find_packages

setup(name=u"canalweb",
      version="0.4",
      author="Frederic Gobry",
      author_email="gobry@yorglu.net",
      license="GPLv3",
      description="Offline viewing of Canal+ shows",
      url="http://github.com/chbug/canalweb",
      packages=find_packages(),
      entry_points={
        'console_scripts': [
            'canalweb = canalweb.cli:main',
            ],
        },
      )



"""VOD-related operations."""

import cookielib
import datetime
import logging
import os
import re
import urllib2
from xml.etree import ElementTree

import canalweb


class VOD(object):  # pylint: disable-msg=R0903
    """Represents a single video."""

    def __init__(self, url, show, vid, stamp):
        self.url = url
        self.show = show
        self.vid = vid
        self.date = stamp
        self.local_path = None
        self.metadata = None

    def __cmp__(self, other):
        return cmp(self.url, other.url)

    def __hash__(self):
        return hash(self.url)

    def __repr__(self):
        return 'VOD(%s, %d)' % (self.show.nickname, self.vid)

    def nice_date(self):
        """Return a nicely formatted date for the show."""
        if self.date:
            return self.date.strftime('%Y-%m-%d')
        return 'nodate'


class Downloader(object):
    """Download videos, possibly using locally cached data."""

    _URL_DOMAIN = 'http://www.canalplus.fr'

    _SHOW_URL = {
        'zapping': '/c-infos-documentaires/pid1830-c-zapping.html',
        'guignoles': '/c-humour/pid1784-c-les-guignols.html',
        'magzin': '/c-humour/pid1787-c-groland.html',
    }

    _CONTENT_RE = re.compile(r'CONTENT_ID.*=\s*["\'](\d+)')

    _CFG_URL = ('http://www.canalplus.fr/flash/xml/module/embed-video-player'
                '/embed-video-player.php?video_id=%(video_id)s&pid=%(pid)d')

    def __init__(self, cachedir):
        self.cachedir = cachedir

    def get_video_info(self, show):
        """Return the list of known videos for a show.

        Args:
          show: str, name of the show
        Returns:
          [VOD()] the known videos
        """
        videos = []
        for vid in self._extract_content_ids(self._fetch_html(show)):
            videos.append(self.fetch_config(show, vid))
        return videos

    def get_cached_configs(self):
        """Return the list of shows locally cached."""
        md_dir = self._get_md_dir()
        videos = {}
        for filename in os.listdir(md_dir):
            parts = os.path.splitext(filename)[0].split('-')
            assert len(parts) == 2
            show, vid = (int(x) for x in parts)
            show = canalweb.SHOW_BY_PID[show]

            full_name = os.path.join(md_dir, filename)
            vod = self._video_url(open(full_name).read(), show, vid)
            vod.metadata = full_name
            videos.setdefault(show, set()).add(vod)
        return videos

    def _fetch_html(self, show):
        """Fetch the HTML page of a show."""
        url = self._URL_DOMAIN + self._SHOW_URL[show.nickname]

        txdata = None
        txheaders = {   
            'User-Agent': 'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US;'
            ' rv:1.8.1.7) Gecko/20070914 Firefox/2.0.0.7'
        }
        cookies = cookielib.CookieJar()
        opener = urllib2.build_opener(urllib2.HTTPCookieProcessor(cookies))
        request = urllib2.Request(url, txdata, txheaders)

        logging.info('fetching %s' % (url,))
        return opener.open(request)

    def _extract_content_ids(self, stream):
        """Extract the show ids from the web page."""
        for line in stream:
            match = self._CONTENT_RE.search(line)
            if not match:
                continue
            yield int(match.group(1))

    def _get_md_dir(self):
        """Return the location of the metadata cache."""
        md_dir = os.path.join(self.cachedir, '.metadata')
        if not os.path.isdir(md_dir):
            os.mkdir(md_dir)
        return md_dir

    def fetch_config(self, show, vid):
        """Get the URL of a video stream, possibly locally cached metadata."""
        md_dir = self._get_md_dir()
        target = os.path.join(md_dir, '%d-%d.xml' % (show.pid, vid))
        if os.path.exists(target):
            data = open(target).read()
        else:
            url = self._CFG_URL % {'video_id': vid, 'pid': show.pid}
            data = urllib2.urlopen(url).read().lstrip()
            open(target, 'w').write(data)

        vod = self._video_url(data, show, vid)
        vod.metadata = target
        return vod

    def _video_url(self, data, show, vid):
        """Parse the XML metadata and extract the video strsm URL."""
        tree = ElementTree.ElementTree(ElementTree.XML(data))
        url = tree.find('//video/hi').text
        # heuristic to get date from filename...
        match = re.search(r'_(\d{6})_', url.split('/')[-1])
        if match:
            info = match.group(1)
            stamp = datetime.datetime(
                year=2000 + int(info[:2]),
                month=int(info[2:4]),
                day=int(info[4:6]))
        else:
            stamp = None
        vod = VOD(url, show, vid, stamp)
        vod.local_path = os.path.join(self.cachedir, '%s-%s-%s.flv' % (
            vod.show.nickname, vod.vid, vod.nice_date()))
        return vod

    @classmethod
    def fetch_video(cls, video):
        """Download a video if it is not locally available."""
        name = video.local_path
        if os.path.exists(name):
            logging.info('%s already available' % (name,))
            return video

        logging.info('downloading %s' % (video.url,))

        filename = name + '.tmp'
        dst = open(filename, 'w')
        src = urllib2.urlopen(video.url)
        while True:
            block = src.read(256 * (2 ** 10))
            dst.write(block)
            if not block:
                break
        dst.close()        
        os.rename(filename, name)
        return video

    @classmethod
    def delete_video(cls, video):
        """Delete metadata and video file."""
        if os.path.exists(video.local_path):
            os.unlink(video.local_path)
        if os.path.exists(video.metadata):
            os.unlink(video.metadata)
